import { ApiPromise, WsProvider } from '@polkadot/api'
import type { Index } from '@polkadot/types/interfaces/runtime'
import { CodecArg, ISubmittableResult } from '@polkadot/types/types'
import { types } from '@joystream/types/'
import { TypeRegistry } from '@polkadot/types'
import { KeyringPair } from '@polkadot/keyring/types'
import { SubmittableExtrinsic } from '@polkadot/api/types'
import {
  DispatchError,
  DispatchResult,
} from '@polkadot/types/interfaces/system'
import { getNonce } from './transactionNonceKeeper'
import logger from '../../services/logger'
import ExitCodes from '../../command-base/ExitCodes'
import { CLIError } from '@oclif/errors'

/**
 * Dedicated error for the failed extrinsics.
 */
export class ExtrinsicFailedError extends CLIError {}

/**
 * Initializes the runtime API and Joystream runtime types.
 *
 * @param apiUrl - API URL string
 * @returns runtime API promise
 */
export async function createApi(apiUrl: string): Promise<ApiPromise> {
  const provider = new WsProvider(apiUrl)

  return await ApiPromise.create({ provider, types })
}

/**
 * Sends an extrinsic to the runtime and follows the result.
 *
 * @param api - API promise
 * @param account - KeyPair instance
 * @param tx - runtime transaction object to send
 * @param nonce - transaction nonce for a given account.
 * @returns extrinsic result promise.
 */
function sendExtrinsic(
  api: ApiPromise,
  account: KeyringPair,
  tx: SubmittableExtrinsic<'promise'>,
  nonce: Index
): Promise<ISubmittableResult> {
  return new Promise((resolve, reject) => {
    let unsubscribe: () => void
    tx.signAndSend(account, { nonce }, (result) => {
      // Implementation loosely based on /pioneer/packages/react-signer/src/Modal.tsx
      if (!result || !result.status) {
        return
      }

      if (result.status.isInBlock) {
        unsubscribe()
        result.events
          .filter(({ event }) => event.section === 'system')
          .forEach(({ event }) => {
            if (event.method === 'ExtrinsicFailed') {
              const dispatchError = event.data[0] as DispatchError
              let errorMsg = dispatchError.toString()
              if (dispatchError.isModule) {
                try {
                  // Need to assert that registry is of TypeRegistry type, since Registry intefrace
                  // seems outdated and doesn't include DispatchErrorModule as possible argument for "findMetaError"
                  const typeRegistry = api.registry as TypeRegistry
                  const { name, documentation } = typeRegistry.findMetaError(
                    dispatchError.asModule
                  )
                  errorMsg = `${name} (${documentation})`
                } catch (e) {
                  // This probably means we don't have this error in the metadata
                  // In this case - continue (we'll just display dispatchError.toString())
                }
              }
              reject(
                new ExtrinsicFailedError(
                  `Extrinsic execution error: ${errorMsg}`,
                  {
                    exit: ExitCodes.ApiError,
                  }
                )
              )
            } else if (event.method === 'ExtrinsicSuccess') {
              const sudid = result.findRecord('sudo', 'Sudid')

              if (sudid) {
                const dispatchSuccess = sudid.event.data[0] as DispatchResult
                if (dispatchSuccess.isOk) {
                  resolve(result)
                } else {
                  reject(
                    new ExtrinsicFailedError(
                      'Sudo extrinsic execution error!',
                      {
                        exit: ExitCodes.ApiError,
                      }
                    )
                  )
                }
              } else {
                resolve(result)
              }
            }
          })
      } else if (result.isError) {
        reject(
          new ExtrinsicFailedError('Extrinsic execution error!', {
            exit: ExitCodes.ApiError,
          })
        )
      }
    })
      .then((unsubFunc) => (unsubscribe = unsubFunc))
      .catch((e) =>
        reject(
          new ExtrinsicFailedError(
            `Cannot send the extrinsic: ${
              e.message ? e.message : JSON.stringify(e)
            }`,
            {
              exit: ExitCodes.ApiError,
            }
          )
        )
      )
  })
}

/**
 * Helper function for sending an extrinsic to the runtime. It constructs an
 * actual transaction object.
 *
 * @param api - API promise
 * @param account - KeyPair instance
 * @param module - runtime module name
 * @param method - runtime extrinsic name
 * @param params - extrinsic parameter
 * @param sudoCall - defines whether the transaction call should be wrapped in
 * the sudo call.
 * @returns void promise.
 */
export async function sendAndFollowNamedTx(
  api: ApiPromise,
  account: KeyringPair,
  module: string,
  method: string,
  params: CodecArg[],
  sudoCall = false
): Promise<void> {
  logger.debug(`Sending ${module}.${method} extrinsic...`)
  let tx = api.tx[module][method](...params)
  if (sudoCall) {
    tx = api.tx.sudo.sudo(tx)
  }
  const nonce = await getNonce(api, account)

  await sendExtrinsic(api, account, tx, nonce)
  logger.debug(`Extrinsic successful!`)
}

/**
 * Helper function for sending an extrinsic to the runtime. It constructs an
 * actual transaction object and sends a transactions wrapped in sudo call.
 *
 * @param api - API promise
 * @param account - KeyPair instance
 * @param module - runtime module name
 * @param method - runtime extrinsic name
 * @param params - extrinsic parameter
 * @returns void promise.
 */
export async function sendAndFollowSudoNamedTx(
  api: ApiPromise,
  account: KeyringPair,
  module: string,
  method: string,
  params: CodecArg[]
): Promise<void> {
  return sendAndFollowNamedTx(api, account, module, method, params, true)
}
